#------------------------------------------------------------------------------
# HATE -- Harness for Algorithm Testing and Evaluation
#------------------------------------------------------------------------------
# See the routine HATE_Help for authorship and copyright details
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# HATE_Args -- process command-line arguments
#------------------------------------------------------------------------------

proc HATE_Args {argv} {
    set inquals 1
    set argc [llength $argv]
    for {set i 0} {$i < $argc} {incr i} {
	set arg [lindex $argv $i]

	if {$inquals && [regexp {^-} $arg]} {
	    if {! [string compare $arg "-detail"]} {
		incr i
		if {$i >= $argc} {
		    HATE_Error "-detail qualifier requires a value"
		}
		HATE_SetDetail [lindex $argv $i]
	    } elseif {! [string compare $arg "-eval"] ||
		! [string compare $arg "-evaluate"]} {
		HATE_SetMode eval
	    } elseif {! [string compare $arg "-help"]} {
		HATE_Help
	    } elseif {! [string compare $arg "-interface"]} {
		incr i
		if {$i >= $argc} {
		    HATE_Error "-interface qualifier requires a value"
		}
		HATE_SetInterface [lindex $argv $i]
	    } elseif {! [string compare $arg "-library"]} {
		incr i
		if {$i >= $argc} {
		    HATE_Error "-library qualifier requires a value"
		}
		HATE_SetLibrary [lindex $argv $i]
	    } elseif {! [string compare $arg "-mode"]} {
		incr i
		if {$i >= $argc} {
		    HATE_Error "-mode qualifier requires a value"
		}
		HATE_SetMode [lindex $argv $i]
	    } elseif {! [string compare $arg "-output"]} {
		incr i
		if {$i >= $argc} {
		    HATE_Error "-output qualifier requires a value"
		}
		HATE_SetGenerate [lindex $argv $i]
	    } elseif {! [string compare $arg "-quiet"]} {
		HATE_SetDetail 1
	    } elseif {! [string compare $arg "-scripts"]} {
		HATE_ListScripts
		exit 1                        ;# SYSTEM-SPECIFIC exit code
	    } elseif {! [string compare $arg "-silent"]} {
		HATE_SetDetail 0
	    } elseif {! [string compare $arg "-test"]} {
		set HATE(mode) test
	    } elseif {! [string compare $arg "-verbose"]} {
		HATE_SetDetail 2
	    } elseif {! [string compare $arg "-version"]} {
		puts stderr [HATE_Version]
	    } elseif {! [string compare $arg "--"]} {
		set inquals 0
	    } else {
		HATE_Error "Unknown qualifier `$arg'" 0
		HATE_Usage
	    }
	} else {
	    HATE_SetScripts $arg
	}
    }

    # Source the interface script and ensure it defines HATE_Interface

    set f [HATE_GetInterface]
    if {[file exists $f]} {
	source $f
    } else {
	HATE_Error "Interface file `$f' doesn't exist."
    }

    if {! [string compare [info command "HATE_Interface"] ""]} {
	HATE_Error \
           "Interface file `$f' doesn't contain procedure `HATE_Interface'"
    }

    # If no script was supplied, generate a usage message

    if {[HATE_GetNumberOfScripts] == 0} {
	HATE_Usage
    }
}


#------------------------------------------------------------------------------
# HATE_BeginOutput -- begin the output for the tests
#------------------------------------------------------------------------------

proc HATE_BeginOutput {} {
    if {[HATE_GetDetail] > 0} {
	set gen [HATE_GetGenerate]
	HATE_BeginOutput_$gen
    }
}


#------------------------------------------------------------------------------
# HATE_BeginOutput_database -- begin the output for the tests for `db'
#------------------------------------------------------------------------------

proc HATE_BeginOutput_database {} {
}


#------------------------------------------------------------------------------
# HATE_BeginOutput_html -- begin the output for the tests for HTML
#------------------------------------------------------------------------------

proc HATE_BeginOutput_html {} {
    HATE_Line "<TABLE>" ""
    HATE_Line "<TR><TH>Script</TH> <TH>No. tests</TH><TH>True Pos</TH>\
	    <TH>True Neg</TH><TH>False Pos</TH><TH>False Neg</TH></TR>"
}


#------------------------------------------------------------------------------
# HATE_BeginOutput_latex -- begin the output for the tests for LaTeX
#------------------------------------------------------------------------------

proc HATE_BeginOutput_latex {} {
    HATE_Line "\\begin\{tabular\}\{|l|r|r|r|r|r|\}" ""
    HATE_Line "\\hline"
    HATE_Line "\{\\bf Script\} & \{\\bf No.~tests\} & \{\\bf True Pos\} &\
	    \{\\bf True Neg\} & \{\\bf False Pos\} & \{\\bf False Neg\}\\\\"
    HATE_Line "\\hline"
}


#------------------------------------------------------------------------------
# HATE_BeginOutput_text -- begin the output for the tests for text
#------------------------------------------------------------------------------

proc HATE_BeginOutput_text {} {
}


#------------------------------------------------------------------------------
# HATE_EndOutput -- end the output for the tests
#------------------------------------------------------------------------------

proc HATE_EndOutput {} {
    if {[HATE_GetDetail] > 0} {
	set gen [HATE_GetGenerate]
	HATE_EndOutput_$gen
    }
}


#------------------------------------------------------------------------------
# HATE_EndOutput_database -- end the output for the tests for `db'
#------------------------------------------------------------------------------

proc HATE_EndOutput_database {} {
}


#------------------------------------------------------------------------------
# HATE_EndOutput_html -- end the output for the tests for HTML
#------------------------------------------------------------------------------

proc HATE_EndOutput_html {} {
    HATE_Line "</TABLE>" ""
}


#------------------------------------------------------------------------------
# HATE_EndOutput_latex -- end the output for the tests for LaTeX
#------------------------------------------------------------------------------

proc HATE_EndOutput_latex {} {
    HATE_Line "\\hline"
    HATE_Line "\\end\{tabular\}" ""
}


#------------------------------------------------------------------------------
# HATE_EndOutput_text -- end the output for the tests for text
#------------------------------------------------------------------------------

proc HATE_EndOutput_text {} {
}


#------------------------------------------------------------------------------
# HATE_Error -- output an error message and die
#------------------------------------------------------------------------------

proc HATE_Error {msg {quit 1}} {
    puts stderr "HATE--error--$msg"
    if {$quit} {
	exit 1                                    ;# SYSTEM-SPECIFIC exit code
    }
}


#------------------------------------------------------------------------------
# HATE_FindScript -- find a script file in the path of script directories
#------------------------------------------------------------------------------

proc HATE_FindScript {s} {
    foreach dir [HATE_GetLibrary] {
	set fn [file join $dir $s]
	if {[file exists $fn]} {
	    return $fn
	}
    }
    return $s
}


#------------------------------------------------------------------------------
# HATE_GetDetail -- get the amount of detail to be reported
#------------------------------------------------------------------------------

proc HATE_GetDetail {} {
    return [HATE_GetVar detail]
}


#------------------------------------------------------------------------------
# HATE_GetGenerate -- get the output format to generate
#------------------------------------------------------------------------------

proc HATE_GetGenerate {} {
    return [HATE_GetVar generate]
}


#------------------------------------------------------------------------------
# HATE_GetInterface -- get the interface script filename
#------------------------------------------------------------------------------

proc HATE_GetInterface {} {
    return [HATE_GetVar interface]
}


#------------------------------------------------------------------------------
# HATE_GetLibrary -- get the library directories
#------------------------------------------------------------------------------

proc HATE_GetLibrary {} {
    return [HATE_GetVar library]
}


#------------------------------------------------------------------------------
# HATE_GetMode -- get the mode of operation
#------------------------------------------------------------------------------

proc HATE_GetMode {} {
    return [HATE_GetVar mode]
}


#------------------------------------------------------------------------------
# HATE_GetNumberOfScripts -- return number of scripts on command line
#------------------------------------------------------------------------------

proc HATE_GetNumberOfScripts {} {
    return [HATE_GetVar scriptSupplied]
}


#------------------------------------------------------------------------------
# HATE_GetScriptName -- get the name of the current test scripts
#------------------------------------------------------------------------------

proc HATE_GetScriptName {} {
    return [HATE_GetVar ScriptName]
}


#------------------------------------------------------------------------------
# HATE_GetScripts -- get the list of test scripts
#------------------------------------------------------------------------------

proc HATE_GetScripts {} {
    return [HATE_GetVar scripts]
}


#------------------------------------------------------------------------------
# HATE_GetVar -- output the value of an element of the HATE global array
#------------------------------------------------------------------------------

proc HATE_GetVar {what} {
    global HATE

    if {[info exists HATE($what)]} {
	return $HATE($what)
    } else {
	return "<unknown>"
    }
}


#------------------------------------------------------------------------------
# HATE_Help -- give some help to the poor user
#------------------------------------------------------------------------------

proc HATE_Help {} {
    puts stderr [HATE_Version] nonewline

    puts stderr "
HATE is an acronym for \"Harness for Algorithm Testing and Evaluation.\"
It may be used for testing an algorithm for correctness (when fixing bugs,
for example) by running regression tests -- tests that exercise different
paths through the code and for which the answers are known precisely.

HATE can also be used in an `evaluate' mode, wherein it accumulates
statistical measures that describe an algorithm's performance.  For
further information, see the World-Wide Web page

   http://peipa.essex.ac.uk/hate/

HATE is copyright 1996 by Adrian F. Clark <alien@essex.ac.uk>."

    puts stderr "\nConfiguration"
    puts stderr "  Output format: [HATE_GetGenerate]"
    puts stderr "  Interface script file: [HATE_GetInterface]"
    puts stderr "  Mode of operation: [HATE_GetMode]"
    puts stderr "  Script library directories: [HATE_GetLibrary]"
    puts stderr "  Verbosity: [HATE_GetDetail]"
    puts stderr ""
}


#------------------------------------------------------------------------------
# HATE_Init -- initialize HATE
#------------------------------------------------------------------------------

proc HATE_Init {} {
    global env

    # Default verbosity level for messages (0--2, 0 being lowest)
    HATE_SetDetail 1

    # Default output format to generate
    HATE_SetGenerate "text"

    # Default name of interface script file
    HATE_SetInterface "interface.tcl"

    # List of directories to search for scripts: initially the current
    # directory, a system-specific default, and any directories in the
    # environment variable "HATE_LIBRARY"
    HATE_SetLibrary "."
    HATE_SetLibrary "/usr/local/lib/hate"     ;# SYSTEM-SPECIFIC library dir

    if {[info exists env(HATE_LIBRARY)]} {
	set lib_path [split $env(HATE_LIBRARY) :]
	foreach dir $lib_path {
	    HATE_SetLibrary $dir
	}
    }

    # Default mode of operation
    HATE_SetMode "test"

    # Clear the variables used for gathering statistics
    HATE_Reset -totals

    # We have no script to process initially
    HATE_SetNumberOfScripts 0

    # Process any initialization file in the user's home directory and
    # in the current directory
    catch {source ~/.hate} junk                    ;# SYSTEM-SPECIFIC
    catch {source .hate} junk                      ;# SYSTEM-SPECIFIC
}


#------------------------------------------------------------------------------
# HATE_Line -- output a line of text
#------------------------------------------------------------------------------

proc HATE_Line {line {indent "   "} {chan stdout}} {
    puts $chan "$indent$line"
}


#------------------------------------------------------------------------------
# HATE_ListScripts -- list the purposes of scripts
#------------------------------------------------------------------------------

proc HATE_ListScripts {} {
    foreach dir [HATE_GetLibrary] {
	if {! [catch {set files [glob [file join $dir *.tcl]]}] } {
	    foreach file $files {
		if {! [catch {source $file}]} {
		    HATE_Line $file ""
		    catch {HATE_DescribeTests}
		    catch {rename HATE_DescribeTests {}}
		    catch {rename HATE_RunTests {}}
		}
	    }
	}
    }
}


#------------------------------------------------------------------------------
# HATE_OutputScript --  output the script name
#------------------------------------------------------------------------------

proc HATE_OutputScript {name} {
    HATE_SetScriptName $name
    if {[HATE_GetDetail] > 0} {
	set gen [HATE_GetGenerate]
	HATE_OutputScript_$gen $name
    }
}


#------------------------------------------------------------------------------
# HATE_OutputScript_database --  output the script name for `db'
#------------------------------------------------------------------------------

proc HATE_OutputScript_database {name} {
    HATE_Line "ScriptName: $name" ""
}


#------------------------------------------------------------------------------
# HATE_OutputScript_html --  output the script name as HTML
#------------------------------------------------------------------------------

proc HATE_OutputScript_html {name} {
}


#------------------------------------------------------------------------------
# HATE_OutputScript_latex --  output the script name as LaTeX
#------------------------------------------------------------------------------

proc HATE_OutputScript_latex {name} {
}


#------------------------------------------------------------------------------
# HATE_OutputScript_text --  output the script name as text
#------------------------------------------------------------------------------

proc HATE_OutputScript_text {name} {
    if {[HATE_GetDetail] > 0} {
	HATE_Line "Script: $name" ""
    }
}


#------------------------------------------------------------------------------
# HATE_Plural -- return a singular or plural version of a phrase
#------------------------------------------------------------------------------

proc HATE_Plural {var text} {
    if {$var == 1} {
	return "$var $text"
    } else {
	return "$var [set text]s"
    }
}


#------------------------------------------------------------------------------
# HATE_Report -- report the result of testing
#------------------------------------------------------------------------------

proc HATE_Report {args} {
    global HATE

    set gen [HATE_GetGenerate]
    set argc [llength $args]
    if {[HATE_GetDetail] > 0} {
	for {set i 0} {$i < $argc} {incr i} {
	    set arg [lindex $args $i]
	    if {! [string compare $arg "-totals"]} {
		HATE_ReportTotals_$gen
	    } elseif {! [string compare $arg "-counters"]} {
		HATE_ReportCounters_$gen
	    } else {
		HATE_Error "unknown qualifier \"$arg\" to HATE_Report"
	    }
	}
    }
}


#------------------------------------------------------------------------------
# HATE_ReportCounters_database -- report the counters as `db'
#------------------------------------------------------------------------------

proc HATE_ReportCounters_database {args} {
    global HATE

    HATE_Line "ScriptSummary: [HATE_GetScriptName] $HATE(count,tests)\
	    $HATE(count,tpos) $HATE(count,tneg) $HATE(count,fpos)\
	    $HATE(count,fneg)" ""
}


#------------------------------------------------------------------------------
# HATE_ReportCounters_html -- report the counters as HTML
#------------------------------------------------------------------------------

proc HATE_ReportCounters_html {args} {
    global HATE

    HATE_Line "<TR><TD><TT>[HATE_GetScriptName]</TT></TD>\
	    <TD>$HATE(count,tests)</TD>\
	    <TD>$HATE(count,tpos)</TD> <TD>$HATE(count,tneg)</TD>\
	    <TD>$HATE(count,fpos)</TD> <TD>$HATE(count,fneg)</TD></TR>"
}


#------------------------------------------------------------------------------
# HATE_ReportCounters_latex -- report the counters as LaTeX
#------------------------------------------------------------------------------

proc HATE_ReportCounters_latex {args} {
    global HATE

    HATE_Line "\{\\tt [HATE_GetScriptName]\} & $HATE(count,tests) &\
	    $HATE(count,tpos) & $HATE(count,tneg) & $HATE(count,fpos) &\
	    $HATE(count,fneg) \\\\"
}


#------------------------------------------------------------------------------
# HATE_ReportCounters_text -- report the counters as text
#------------------------------------------------------------------------------

proc HATE_ReportCounters_text {args} {
    global HATE

    HATE_Line "[HATE_Plural $HATE(count,tests) test] were executed. There were\
	    [HATE_Plural $HATE(count,tpos) {true positive}],\
	    [HATE_Plural $HATE(count,tneg) {true negative}]"
    HATE_Line "[HATE_Plural $HATE(count,fpos) {false positive}], and\
	    [HATE_Plural $HATE(count,fneg) {false negative}]."
}


#------------------------------------------------------------------------------
# HATE_ReportTest -- output result of a single test
#------------------------------------------------------------------------------

proc HATE_ReportTest {name expect input output status} {
    set gen [HATE_GetGenerate]
    HATE_ReportTest_$gen $name $expect $input $output $status
}


#------------------------------------------------------------------------------
# HATE_ReportTest_database -- report the result of a single test as `db'
#------------------------------------------------------------------------------

proc HATE_ReportTest_database {name expect input output status} {
    HATE_Line "Result: \{$name\} \{$status\} \{$input\} \{$output\}" ""
}


#------------------------------------------------------------------------------
# HATE_ReportTest_html -- report the result of a single test as HTML
#------------------------------------------------------------------------------

proc HATE_ReportTest_html {name expect input output status} {
    HATE_Line \
      "<TR><TD>$name</TD><TD>$status</TD><TD>$input</TD><TD>$output</TD></TR>"
}


#------------------------------------------------------------------------------
# HATE_ReportTest_latex -- report the result of a single test as LaTeX
#------------------------------------------------------------------------------

proc HATE_ReportTest_latex {name expect input output status} {
    HATE_Line "$name & $status & \\multicolumn4\{l|\}\
	    \{input \"$input\" output \"$output\"\}\\\\"
}


#------------------------------------------------------------------------------
# HATE_ReportTest_text -- report the result of a single test as text
#------------------------------------------------------------------------------

proc HATE_ReportTest_text {name expect input output status} {
    HATE_Line "$name: $status -- input \"$input\" output \"$output\""
}


#------------------------------------------------------------------------------
# HATE_ReportTotals_database -- report the totals as `db'
#------------------------------------------------------------------------------

proc HATE_ReportTotals_database {args} {
    global HATE

    HATE_Line "Overall: $HATE(total,tests) $HATE(total,tpos)\
	    $HATE(total,tneg) $HATE(total,fpos) $HATE(total,fneg)" ""
}


#------------------------------------------------------------------------------
# HATE_ReportTotals_html -- report the totals as HTML
#------------------------------------------------------------------------------

proc HATE_ReportTotals_html {args} {
    global HATE

}


#------------------------------------------------------------------------------
# HATE_ReportTotals_latex -- report the totals as LaTeX
#------------------------------------------------------------------------------

proc HATE_ReportTotals_latex {args} {
    global HATE

    HATE_Line "\\hline"
    HATE_Line "\\hline"
    HATE_Line "\{\\bf Overall:\} & $HATE(count,tests) & $HATE(count,tpos)\
	    & $HATE(count,tneg) & $HATE(count,fpos) & $HATE(count,fneg) \\\\"
    HATE_Line "\\hline"
}


#------------------------------------------------------------------------------
# HATE_ReportTotals_text -- report the totals as text
#------------------------------------------------------------------------------

proc HATE_ReportTotals_text {args} {
    global HATE

    HATE_Line "Overall results:" ""
    HATE_Line "Number of tests: $HATE(total,tests)"
    HATE_Line "True positives:  $HATE(total,tpos)"
    HATE_Line "True negatives:  $HATE(total,tneg)"
    HATE_Line "False positives: $HATE(total,fpos)"
    HATE_Line "False negatives: $HATE(total,fneg)"
}


#------------------------------------------------------------------------------
# HATE_Reset --  reset things before running a test
#------------------------------------------------------------------------------

proc HATE_Reset {args} {
    global HATE

    set argc [llength $args]
    for {set i 0} {$i < $argc} {incr i} {
	set arg [lindex $args $i]
	if {! [string compare $arg "-totals"]} {
	    set HATE(total,tests) 0
	    set HATE(total,tpos) 0
	    set HATE(total,tneg) 0
	    set HATE(total,fpos) 0
	    set HATE(total,fneg) 0
	} elseif {! [string compare $arg "-counters"]} {
	    set HATE(count,tests) 0
	    set HATE(count,tpos) 0
	    set HATE(count,tneg) 0
	    set HATE(count,fpos) 0
	    set HATE(count,fneg) 0
	} else {
	    HATE_Error "unknown qualifier \"$arg\" to HATE_Reset"
	}
    }
}


#------------------------------------------------------------------------------
# HATE_RunScript -- execute a test script
#------------------------------------------------------------------------------

proc HATE_RunScript {script} {
    # Find the script in the library directories
    set s [HATE_FindScript $script]
    if {[file exists $s]} {
	HATE_OutputScript $s

	# Delete any existing procedure and source the script file
	catch {info command "HATE_RunTests"} msg
	if {$msg == 1} {
	    rename HATE_RunTests {}
	}
	source $s

	# Ensure the script file contained the procedure and run it
	if {[string compare [info command "HATE_RunTests"] ""]} {
	    HATE_Reset -counters
	    HATE_RunTests
	} else {
	    HATE_Error \
               "Script file `$s' doesn't contain a procedure `HATE_RunTests'"
	}
    } else {
	HATE_Error "Script file `$script' not found in HATE script path."
    }
}


#------------------------------------------------------------------------------
# HATE_SetDetail -- set the amount of detail to be reported
#------------------------------------------------------------------------------

proc HATE_SetDetail {d} {
    global HATE

    set HATE(detail) $d
}


#------------------------------------------------------------------------------
# HATE_SetGenerate -- set the output format to generate
#------------------------------------------------------------------------------

proc HATE_SetGenerate {f} {
    global HATE

    if {![string compare $f "latex"] || ![string compare $f "LaTeX"]} {
	set HATE(generate) latex
    } elseif {![string compare $f "html"] || ![string compare $f "HTML"]} {
	set HATE(generate) html
    } elseif {![string compare $f "database"] || ![string compare $f "db"]} {
	set HATE(generate) database
	HATE_SetDetail 2
    } elseif {![string compare $f "text"]} {
	set HATE(generate) text
    } else {
	HATE_Error "Unknown output format `$f' requested."
    }
}


#------------------------------------------------------------------------------
# HATE_SetInterface -- set the name of the file containing the interface script
#------------------------------------------------------------------------------

proc HATE_SetInterface {f} {
    global HATE

    set HATE(interface) $f
}


#------------------------------------------------------------------------------
# HATE_SetLibrary -- set the directories in which to look for library files
#------------------------------------------------------------------------------

proc HATE_SetLibrary {l} {
    global HATE

    lappend HATE(library) $l
}


#------------------------------------------------------------------------------
# HATE_SetMode -- set HATE into "eval" or "test" mode
#------------------------------------------------------------------------------

proc HATE_SetMode {mode} {
    global HATE

    if {! [string compare $mode "eval"]} {
	set HATE(mode) eval
    } elseif {! [string compare $mode "test"]} {
	set HATE(mode) test
    } else {
	HATE_Error "Attempt to set unknown \"$mode\" mode!"
    }
}


#------------------------------------------------------------------------------
# HATE_SetNumberOfScripts -- set the "script supplied" flag
#------------------------------------------------------------------------------

proc HATE_SetNumberOfScripts {l} {
    global HATE

    if {$l == 0} {
	set HATE(scriptSupplied) $l
    } else {
	incr HATE(scriptSupplied) $l
    }
}


#------------------------------------------------------------------------------
# HATE_SetScriptName -- record the name of the current script
#------------------------------------------------------------------------------

proc HATE_SetScriptName {s} {
    global HATE

    set HATE(ScriptName) $s
}


#------------------------------------------------------------------------------
# HATE_SetScripts -- add a test script to the list of scripts
#------------------------------------------------------------------------------

proc HATE_SetScripts {s} {
    global HATE

    lappend HATE(scripts) $s
    HATE_SetNumberOfScripts 1
}


#------------------------------------------------------------------------------
# HATE_Test -- run a single test
#------------------------------------------------------------------------------

proc HATE_Test {name expect input output} {
    global HATE

    set result [eval [list HATE_Interface $input]]
    if {! [string compare $output $result]} {
	# The result matches the expected output
	if {$expect} {
	    incr HATE(count,tpos)
	    incr HATE(total,tpos)
	    set OK 1
	} else {
	    incr HATE(count,fpos)
	    incr HATE(total,fpos)
	    set OK 0
	}
    } else {
	# The result doesn't match the expected output
	if {$expect} {
	    incr HATE(count,fneg)
	    incr HATE(total,fneg)
	    set OK 0
	} else {
	    incr HATE(count,tneg)
	    incr HATE(total,tneg)
	    set OK 1
	}
    }
    incr HATE(count,tests)
    incr HATE(total,tests)

    # If we have an unexpected result, we might produce a warning

    if {$OK} {
	set status "success"
    } else {
	set status "failure"
    }
    set detail [HATE_GetDetail]
    if {! $OK && ($detail > 0) && ![string compare [HATE_GetMode] "test"]} {
	HATE_ReportTest $name $expect $input $output $status
    } elseif {$detail > 1} {
	HATE_ReportTest $name $expect $input $output $status
    }
}


#------------------------------------------------------------------------------
# HATE_Usage -- output a usage message and exit
#------------------------------------------------------------------------------

proc HATE_Usage {} {
    puts stderr "Usage: [file tail [info script]] \[-qualifiers\] <script>"
    puts stderr "where"
    puts stderr "  -detail <n>                         set level of detail "
    puts stderr "  -eval\[uate\]                         as \"-mode eval\""
    puts stderr "  -help                               give help on HATE"
    puts stderr "  -interface <file>                   program interface file"
    puts stderr "  -library <dir>                      add test directory"
    puts stderr "  -mode test|eval                     set mode of operation"
    puts stderr "  -output html|latex|database|text    specify output format"
    puts stderr "  -quiet                              as \"-detail 1\""
    puts stderr "  -silent                             as \"-detail 0\""
    puts stderr "  -test                               as \"-mode test\""
    puts stderr "  -verbose                            as \"-detail 2\""
    puts stderr "  -version                            output HATE version"
    exit 1                                     ;# SYSTEM-SPECIFIC exit code
}


#------------------------------------------------------------------------------
# HATE_Version -- return the version string
#------------------------------------------------------------------------------

proc HATE_Version {} {
    puts stderr "HATE version 0.00 beta"
}


#------------------------------------------------------------------------------
# HATE_Warn -- output an error message
#------------------------------------------------------------------------------

proc HATE_Warn {msg} {
    puts stderr "HATE--warning--$msg"
}
